@@ -8,8 +8,10 @@ import Control.Exception
 import Control.Monad
 import Control.Monad.IO.Class
 
+import Data.Containers.ListUtils
 import Data.Either
 import Data.List
+import Data.Maybe
 import Data.Text (Text)
 import Data.Text qualified as T
 import Data.Text.IO qualified as T
@@ -22,6 +24,8 @@ import Command
 import Config
 import Eval
 import Job
+import Job.Types
+import Output
 import Repo
 import Terminal
 
@@ -29,12 +33,19 @@ import Terminal
 data RunCommand = RunCommand RunOptions [ Text ]
 
 data RunOptions = RunOptions
-    { roRanges :: [ Text ]
+    { roRerun :: RerunOption
+    , roRanges :: [ Text ]
     , roSinceUpstream :: [ Text ]
     , roNewCommitsOn :: [ Text ]
     , roNewTags :: [ Pattern ]
     }
 
+data RerunOption
+    = RerunExplicit
+    | RerunFailed
+    | RerunAll
+    | RerunNone
+
 instance Command RunCommand where
     commandName _ = "run"
     commandDescription _ = "Execude jobs per minici.yaml for given commits"
@@ -54,14 +65,27 @@ instance Command RunCommand where
 
     type CommandOptions RunCommand = RunOptions
     defaultCommandOptions _ = RunOptions
-        { roRanges = []
+        { roRerun = RerunExplicit
+        , roRanges = []
         , roSinceUpstream = []
         , roNewCommitsOn = []
         , roNewTags = []
         }
 
     commandOptions _ =
-        [ Option [] [ "range" ]
+        [ Option [] [ "rerun-explicit" ]
+            (NoArg (\opts -> opts { roRerun = RerunExplicit }))
+            "rerun jobs given explicitly on command line and their failed dependencies (default)"
+        , Option [] [ "rerun-failed" ]
+            (NoArg (\opts -> opts { roRerun = RerunFailed }))
+            "rerun failed jobs only"
+        , Option [] [ "rerun-all" ]
+            (NoArg (\opts -> opts { roRerun = RerunAll }))
+            "rerun all jobs"
+        , Option [] [ "rerun-none" ]
+            (NoArg (\opts -> opts { roRerun = RerunNone }))
+            "do not rerun any job"
+        , Option [] [ "range" ]
             (ReqArg (\val opts -> opts { roRanges = T.pack val : roRanges opts }) "<range>")
             "run jobs for commits in given range"
         , Option [] [ "since-upstream" ]
@@ -123,26 +147,80 @@ mergeSources sources = do
 argumentJobSource :: [ JobName ] -> CommandExec JobSource
 argumentJobSource [] = emptyJobSource
 argumentJobSource names = do
-    config <- getConfig
-    einput <- getEvalInput
-    jobsetJobsEither <- fmap Right $ forM names $ \name ->
+    jobRoot <- getJobRoot
+    ( config, jcommit ) <- case jobRoot of
+        JobRootConfig config -> do
+            commit <- sequence . fmap createWipCommit =<< tryGetDefaultRepo
+            return ( config, commit )
+        JobRootRepo repo -> do
+            commit <- createWipCommit repo
+            config <- either fail return =<< loadConfigForCommit =<< getCommitTree commit
+            return ( config, Just commit )
+
+    jobtree <- case jcommit of
+        Just commit -> (: []) <$> getCommitTree commit
+        Nothing -> return []
+    let cidPart = case jobRoot of
+            JobRootConfig {} -> []
+            JobRootRepo {} -> map (JobIdTree Nothing "" . treeId) jobtree
+    forM_ names $ \name ->
         case find ((name ==) . jobName) (configJobs config) of
-            Just job -> return job
-            Nothing -> tfail $ "job `" <> textJobName name <> "' not found"
-    jobsetCommit <- sequence . fmap createWipCommit =<< tryGetDefaultRepo
-    oneshotJobSource [ evalJobSet einput JobSet {..} ]
+            Just _  -> return ()
+            Nothing -> tfail $ "job ‘" <> textJobName name <> "’ not found"
+
+    jset <- cmdEvalWith (\ei -> ei { eiCurrentIdRev = cidPart ++ eiCurrentIdRev ei }) $ do
+        evalJobSetSelected names (map ( Nothing, ) jobtree) JobSet
+            { jobsetId = ()
+            , jobsetConfig = Just config
+            , jobsetCommit = jcommit
+            , jobsetExplicitlyRequested = names
+            , jobsetJobsEither = Right (configJobs config)
+            }
+    oneshotJobSource [ jset ]
+
+refJobSource :: [ JobRef ] -> CommandExec JobSource
+refJobSource [] = emptyJobSource
+refJobSource refs = do
+    sets <- foldl' addJobToList [] <$> cmdEvalWith id (mapM evalJobReferenceToSet refs)
+    oneshotJobSource sets
+  where
+    addJobToList :: [ JobSet ] -> JobSet -> [ JobSet ]
+    addJobToList (cur : rest) jset
+        | jobsetId cur == jobsetId jset = cur { jobsetJobsEither = fmap (nubOrdOn jobId) $ (++) <$> (jobsetJobsEither cur) <*> (jobsetJobsEither jset)
+                                              , jobsetExplicitlyRequested = nubOrd $ jobsetExplicitlyRequested cur ++ jobsetExplicitlyRequested jset
+                                              } : rest
+        | otherwise                     = cur : addJobToList rest jset
+    addJobToList [] jset                = [ jset ]
+
+loadJobSetFromRoot :: (MonadIO m, MonadFail m) => JobRoot -> Commit -> m DeclaredJobSet
+loadJobSetFromRoot root commit = case root of
+    JobRootRepo _ -> loadJobSetForCommit commit
+    JobRootConfig config -> return JobSet
+        { jobsetId = ()
+        , jobsetConfig = Just config
+        , jobsetCommit = Just commit
+        , jobsetExplicitlyRequested = []
+        , jobsetJobsEither = Right $ configJobs config
+        }
 
 rangeSource :: Text -> Text -> CommandExec JobSource
 rangeSource base tip = do
+    root <- getJobRoot
     repo <- getDefaultRepo
-    einput <- getEvalInput
     commits <- listCommits repo (base <> ".." <> tip)
-    oneshotJobSource . map (evalJobSet einput) =<< mapM loadJobSetForCommit commits
+    jobsets <- forM commits $ \commit -> do
+        tree <- getCommitTree commit
+        cmdEvalWith (\ei -> ei
+            { eiCurrentIdRev = JobIdTree Nothing (treeSubdir tree) (treeId tree) : eiCurrentIdRev ei
+            }) . evalJobSet [ ( Nothing, tree) ] =<< loadJobSetFromRoot root commit
+    oneshotJobSource jobsets
+
 
 watchBranchSource :: Text -> CommandExec JobSource
 watchBranchSource branch = do
+    root <- getJobRoot
     repo <- getDefaultRepo
-    einput <- getEvalInput
+    einputBase <- getEvalInput
     getCurrentTip <- watchBranch repo branch
     let go prev tmvar = do
             cur <- atomically $ do
@@ -153,7 +231,13 @@ watchBranchSource branch = do
                     Nothing -> retry
 
             commits <- listCommits repo (textCommitId (commitId prev) <> ".." <> textCommitId (commitId cur))
-            jobsets <- map (evalJobSet einput) <$> mapM loadJobSetForCommit commits
+            jobsets <- forM commits $ \commit -> do
+                tree <- getCommitTree commit
+                let einput = einputBase
+                        { eiCurrentIdRev = JobIdTree Nothing (treeSubdir tree) (treeId tree) : eiCurrentIdRev einputBase
+                        }
+                either (fail . T.unpack . textEvalError) return =<<
+                    flip runEval einput . evalJobSet [ ( Nothing, tree ) ] =<< loadJobSetFromRoot root commit
             nextvar <- newEmptyTMVarIO
             atomically $ putTMVar tmvar $ Just ( jobsets, JobSource nextvar )
             go cur nextvar
@@ -164,20 +248,26 @@ watchBranchSource branch = do
             Just commit ->
                 void $ forkIO $ go commit tmvar
             Nothing -> do
-                T.hPutStrLn stderr $ "Branch `" <> branch <> "' not found"
+                T.hPutStrLn stderr $ "Branch ‘" <> branch <> "’ not found"
                 atomically $ putTMVar tmvar Nothing
         return $ JobSource tmvar
 
 watchTagSource :: Pattern -> CommandExec JobSource
 watchTagSource pat = do
+    root <- getJobRoot
     chan <- watchTags =<< getDefaultRepo
-    einput <- getEvalInput
+    einputBase <- getEvalInput
 
     let go tmvar = do
             tag <- atomically $ readTChan chan
             if match pat $ T.unpack $ tagTag tag
               then do
-                jobset <- evalJobSet einput <$> loadJobSetForCommit (tagObject tag)
+                tree <- getCommitTree $ tagObject tag
+                let einput = einputBase
+                        { eiCurrentIdRev = JobIdTree Nothing (treeSubdir tree) (treeId tree) : eiCurrentIdRev einputBase
+                        }
+                jobset <- either (fail . T.unpack . textEvalError) return =<<
+                    flip runEval einput . evalJobSet [ ( Nothing, tree ) ] =<< loadJobSetFromRoot root (tagObject tag)
                 nextvar <- newEmptyTMVarIO
                 atomically $ putTMVar tmvar $ Just ( [ jobset ], JobSource nextvar )
                 go nextvar
@@ -192,33 +282,41 @@ watchTagSource pat = do
 cmdRun :: RunCommand -> CommandExec ()
 cmdRun (RunCommand RunOptions {..} args) = do
     CommonOptions {..} <- getCommonOptions
-    tout <- getTerminalOutput
+    output <- getOutput
     storageDir <- getStorageDir
 
     ( rangeOptions, jobOptions ) <- partitionEithers . concat <$> sequence
         [ forM roRanges $ \range -> case T.splitOn ".." range of
-            [ base, tip ] -> return $ Left ( Just base, tip )
+            [ base, tip ]
+                | not (T.null base) && not (T.null tip)
+                -> return $ Left ( Just base, tip )
             _ -> tfail $ "invalid range: " <> range
         , forM roSinceUpstream $ return . Left . ( Nothing, )
         , forM args $ \arg -> case T.splitOn ".." arg of
-            [ base, tip ] -> return $ Left ( Just base, tip )
-            [ _ ] -> do
-                config <- getConfig
-                if any ((JobName arg ==) . jobName) (configJobs config)
-                  then return $ Right $ JobName arg
-                  else do
-                    liftIO $ T.hPutStrLn stderr $ "standalone `" <> arg <> "' argument deprecated, use `--since-upstream=" <> arg <> "' instead"
-                    return $ Left ( Nothing, arg )
+            [ base, tip ]
+                | not (T.null base) && not (T.null tip)
+                -> return $ Left ( Just base, tip )
+            [ _ ] -> return $ Right arg
             _ -> tfail $ "invalid argument: " <> arg
         ]
 
-    argumentJobs <- argumentJobSource jobOptions
+    let ( refOptions, nameOptions ) = partition (T.elem '.') jobOptions
 
-    let rangeOptions'
-            | null rangeOptions, null roNewCommitsOn, null roNewTags, null jobOptions = [ ( Nothing, "HEAD" ) ]
-            | otherwise = rangeOptions
+    argumentJobs <- argumentJobSource $ map JobName nameOptions
+    refJobs <- refJobSource $ map parseJobRef refOptions
 
-    ranges <- forM rangeOptions' $ \( mbBase, paramTip ) -> do
+    defaultSource <- getJobRoot >>= \case
+        _ | not (null rangeOptions && null roNewCommitsOn && null roNewTags && null jobOptions) -> do
+            emptyJobSource
+
+        JobRootRepo repo -> do
+            Just base <- findUpstreamRef repo "HEAD"
+            rangeSource base "HEAD"
+
+        JobRootConfig config -> do
+            argumentJobSource (jobName <$> configJobs config)
+
+    ranges <- forM rangeOptions $ \( mbBase, paramTip ) -> do
         ( base, tip ) <- case mbBase of
             Just base -> return ( base, paramTip )
             Nothing -> do
@@ -232,14 +330,16 @@ cmdRun (RunCommand RunOptions {..} args) = do
     liftIO $ do
         mngr <- newJobManager storageDir optJobs
 
-        source <- mergeSources $ concat [ [ argumentJobs ], ranges, branches, tags ]
-        headerLine <- newLine tout ""
+        source <- mergeSources $ concat [ [ defaultSource, argumentJobs, refJobs ], ranges, branches, tags ]
+        mbHeaderLine <- mapM (flip newLine "") (outputTerminal output)
 
         threadCount <- newTVarIO (0 :: Int)
         let changeCount f = atomically $ do
                 writeTVar threadCount . f =<< readTVar threadCount
-        let waitForJobs = atomically $ do
-                flip when retry . (0 <) =<< readTVar threadCount
+        let waitForJobs = do
+                atomically $ do
+                    flip when retry . (0 <) =<< readTVar threadCount
+                waitForRemainingTasks mngr
 
         let loop _ Nothing = return ()
             loop names (Just ( [], next )) = do
@@ -248,9 +348,10 @@ cmdRun (RunCommand RunOptions {..} args) = do
             loop pnames (Just ( jobset : rest, next )) = do
                 let names = nub $ (pnames ++) $ map jobName $ jobsetJobs jobset
                 when (names /= pnames) $ do
-                    redrawLine headerLine $ T.concat $
-                        T.replicate (8 + 50) " " :
-                        map ((" " <>) . fitToLength 7 . textJobName) names
+                    forM_ mbHeaderLine $ \headerLine -> do
+                        redrawLine headerLine $ T.concat $
+                            T.replicate (8 + 50) " " :
+                            map ((" " <>) . fitToLength 7 . textJobName) names
 
                 let commit = jobsetCommit jobset
                     shortCid = T.pack $ take 7 $ maybe (repeat ' ') (showCommitId . commitId) commit
@@ -258,23 +359,34 @@ cmdRun (RunCommand RunOptions {..} args) = do
 
                 case jobsetJobsEither jobset of
                     Right jobs -> do
-                        outs <- runJobs mngr tout commit jobs
+                        outs <- runJobs mngr output jobs $ case roRerun of
+                            RerunExplicit -> \jid status -> jid `elem` jobsetExplicitlyRequested jobset || jobStatusFailed status
+                            RerunFailed -> \_ status -> jobStatusFailed status
+                            RerunAll -> \_ _ -> True
+                            RerunNone -> \_ _ -> False
                         let findJob name = snd <$> find ((name ==) . jobName . fst) outs
-                        line <- newLine tout ""
+                            statuses = map findJob names
+                        forM_ (outputTerminal output) $ \tout -> do
+                            line <- newLine tout ""
+                            void $ forkIO $ do
+                                displayStatusLine tout line shortCid (" " <> shortDesc) statuses
                         mask $ \restore -> do
                             changeCount (+ 1)
-                            void $ forkIO $ (>> changeCount (subtract 1)) $
-                                try @SomeException $ restore $ do
-                                    displayStatusLine tout line shortCid (" " <> shortDesc) $ map findJob names
+                            void $ forkIO $ do
+                                void $ try @SomeException $ restore $ waitForJobStatuses statuses
+                                changeCount (subtract 1)
                     Left err -> do
-                        void $ newLine tout $
+                        forM_ (outputTerminal output) $ flip newLine $
                             "\ESC[91m" <> shortCid <> "\ESC[0m" <> " " <> shortDesc <> " \ESC[91m" <> T.pack err <> "\ESC[0m"
+                        outputEvent output $ TestMessage $ "jobset-fail " <> T.pack err
+                        outputEvent output $ LogMessage $ "Jobset failed: " <> shortCid <> " " <> T.pack err
                 loop names (Just ( rest, next ))
 
         handle @SomeException (\_ -> cancelAllJobs mngr) $ do
             loop [] =<< atomically (takeJobSource source)
             waitForJobs
         waitForJobs
+    outputEvent output $ TestMessage "run-finish"
 
 
 fitToLength :: Int -> Text -> Text
@@ -284,22 +396,26 @@ fitToLength maxlen str | len <= maxlen = str <> T.replicate (maxlen - len) " "
 
 showStatus :: Bool -> JobStatus a -> Text
 showStatus blink = \case
-    JobQueued       -> "\ESC[94m…\ESC[0m      "
+    JobQueued       -> " \ESC[94m…\ESC[0m     "
     JobWaiting uses -> "\ESC[94m~" <> fitToLength 6 (T.intercalate "," (map textJobName uses)) <> "\ESC[0m"
-    JobSkipped      ->  "\ESC[0m-\ESC[0m      "
-    JobRunning      -> "\ESC[96m" <> (if blink then "*" else "•") <> "\ESC[0m      "
-    JobError fnote  -> "\ESC[91m" <> fitToLength 7 ("!! [" <> T.pack (show (footnoteNumber fnote)) <> "]") <> "\ESC[0m"
-    JobFailed       -> "\ESC[91m✗\ESC[0m      "
-    JobCancelled    ->  "\ESC[0mC\ESC[0m      "
-    JobDone _       -> "\ESC[92m✓\ESC[0m      "
+    JobSkipped      ->  " \ESC[0m-\ESC[0m     "
+    JobRunning      -> " \ESC[96m" <> (if blink then "*" else "•") <> "\ESC[0m     "
+    JobError fnote  -> "\ESC[91m" <> fitToLength 7 ("!! [" <> T.pack (maybe "?" (show . tfNumber) (footnoteTerminal fnote)) <> "]") <> "\ESC[0m"
+    JobFailed       -> " \ESC[91m✗\ESC[0m     "
+    JobCancelled    ->  " \ESC[0mC\ESC[0m     "
+    JobDone _       -> " \ESC[92m✓\ESC[0m     "
 
     JobDuplicate _ s -> case s of
-        JobQueued    -> "\ESC[94m^\ESC[0m      "
-        JobWaiting _ -> "\ESC[94m^\ESC[0m      "
-        JobSkipped   ->  "\ESC[0m-\ESC[0m      "
-        JobRunning   -> "\ESC[96m" <> (if blink then "*" else "^") <> "\ESC[0m      "
+        JobQueued    -> " \ESC[94m^\ESC[0m     "
+        JobWaiting _ -> " \ESC[94m^\ESC[0m     "
+        JobSkipped   ->  " \ESC[0m-\ESC[0m     "
+        JobRunning   -> " \ESC[96m" <> (if blink then "*" else "^") <> "\ESC[0m     "
         _            -> showStatus blink s
 
+    JobPreviousStatus (JobDone _) -> "\ESC[90m«\ESC[32m✓\ESC[0m     "
+    JobPreviousStatus (JobFailed) -> "\ESC[90m«\ESC[31m✗\ESC[0m     "
+    JobPreviousStatus s           -> "\ESC[90m«" <> T.init (showStatus blink s)
+
 displayStatusLine :: TerminalOutput -> TerminalLine -> Text -> Text -> [ Maybe (TVar (JobStatus JobOutput)) ] -> IO ()
 displayStatusLine tout line prefix1 prefix2 statuses = do
     go "\0"
@@ -320,3 +436,10 @@ displayStatusLine tout line prefix1 prefix2 statuses = do
         if all (maybe True jobStatusFinished) ss
            then return ()
            else go cur
+
+waitForJobStatuses :: [ Maybe (TVar (JobStatus a)) ] -> IO ()
+waitForJobStatuses mbstatuses = do
+    let statuses = catMaybes mbstatuses
+    atomically $ do
+        ss <- mapM readTVar statuses
+        when (any (not . jobStatusFinished) ss) retry
